package one.press.prsutilityplugin;

import com.fasterxml.jackson.databind.ObjectMapper;
import one.press.prsutilityplugin.args.ArgsValid;
import one.press.prsutilityplugin.args.Pbkdf2Args;

import org.bouncycastle.crypto.generators.PKCS5S2ParametersGenerator;
import org.bouncycastle.crypto.digests.SHA256Digest;
import org.bouncycastle.crypto.params.KeyParameter;

import static java.nio.charset.StandardCharsets.UTF_8;

import java.util.Map;

import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.MethodChannel.MethodCallHandler;
import io.flutter.plugin.common.MethodChannel.Result;
import io.flutter.plugin.common.PluginRegistry.Registrar;

/**
 * PrsUtilityPlugin
 */
public class PrsUtilityPlugin implements MethodCallHandler {

  private ObjectMapper objectMapper = new ObjectMapper();

  /**
   * Plugin registration.
   */
  public static void registerWith(Registrar registrar) {
    final MethodChannel channel = new MethodChannel(registrar.messenger(), "one.press.utility/prs_utility_plugin");
    channel.setMethodCallHandler(new PrsUtilityPlugin());
  }

  @Override
  public void onMethodCall(MethodCall call, Result result) {
    if (call == null || call.method == null) {
      result.error(ErrorCode.CALL_ERROR, "call or call.method is null", call);
      return;
    }
    switch (call.method) {
    case CallMethod.pbkdf2:
      onPbkdf2(call, result);
    default:
      result.notImplemented();
      break;
    }
  }

  private void onPbkdf2(MethodCall call, Result result) {
    try {
      if (isArgumentIllegal(call, result)) {
        return;
      }
      Object arguments = call.arguments;
      String json = objectMapper.writeValueAsString(arguments);
      Pbkdf2Args args = objectMapper.readValue(json, Pbkdf2Args.class);
      if (isArgumentValid(args, call, result)) {
        return;
      }

      PKCS5S2ParametersGenerator gen = new PKCS5S2ParametersGenerator(new SHA256Digest());
      gen.init(args.password.getBytes("UTF-8"), hexStringToByteArray(args.salt), args.iterations);
      byte[] keyBytes = ((KeyParameter) gen.generateDerivedParameters(256)).getKey();
      String key = toHexString(keyBytes, 0, keyBytes.length, false);
      result.success(key);
    } catch (Exception e) {
      e.printStackTrace();
      result.error(ErrorCode.ERROR, e.getMessage(), null);
    }
  }

  public static boolean containsHexPrefix(String input) {
    return input.length() > 1 && input.charAt(0) == '0' && input.charAt(1) == 'x';
  }

  public static String cleanHexPrefix(String input) {
    if (containsHexPrefix(input)) {
      return input.substring(2);
    } else {
      return input;
    }
  }

  public static byte[] hexStringToByteArray(String input) {
    String cleanInput = cleanHexPrefix(input);

    int len = cleanInput.length();

    if (len == 0) {
      return new byte[] {};
    }

    byte[] data;
    int startIdx;
    if (len % 2 != 0) {
      data = new byte[(len / 2) + 1];
      data[0] = (byte) Character.digit(cleanInput.charAt(0), 16);
      startIdx = 1;
    } else {
      data = new byte[len / 2];
      startIdx = 0;
    }

    for (int i = startIdx; i < len; i += 2) {
      data[(i + 1) / 2] = (byte) ((Character.digit(cleanInput.charAt(i), 16) << 4)
          + Character.digit(cleanInput.charAt(i + 1), 16));
    }
    return data;
  }

  public static String toHexString(byte[] input, int offset, int length, boolean withPrefix) {
    StringBuilder stringBuilder = new StringBuilder();
    if (withPrefix) {
      stringBuilder.append("0x");
    }
    for (int i = offset; i < offset + length; i++) {
      stringBuilder.append(String.format("%02x", input[i] & 0xFF));
    }

    return stringBuilder.toString();
  }

  private boolean isArgumentIllegal(MethodCall call, Result result) {
    if (!(call.arguments instanceof Map)) {
      result.error(ErrorCode.ARGS_ERROR, String.format("arguments in %s method type error.need map", call.method),
          null);
      return true;
    }
    return false;
  }

  private boolean isArgumentValid(ArgsValid args, MethodCall call, Result result) {
    if (!args.isValid()) {
      result.error(ErrorCode.ARGS_ERROR, String.format("arguments in %s method param error", call.method), null);
      return true;
    }
    return false;
  }

}
